#include "guecs/ui.hpp"
#include <typeinfo>

namespace guecs {
  using std::make_shared;

  UI::UI() {
    $font = make_shared<sf::Font>(THEME.FONT_FILE_NAME);
  }

  void UI::position(int x, int y, int width, int height) {
    $parser.position(x, y, width, height);
  }

  sf::Vector2f UI::get_position() {
    return {float($parser.grid_x), float($parser.grid_y)};
  }

  sf::Vector2f UI::get_size() {
    return {float($parser.grid_w), float($parser.grid_h)};
  }

  void UI::layout(const string& grid) {
    $grid = grid;
    bool good = $parser.parse($grid);
    dbc::check(good, "LEL parsing failed.");

    for(auto& [name, cell] : $parser.cells) {
      auto ent = init_entity(name);
      set<lel::Cell>(ent, cell);
    }
  }

  Entity UI::init_entity(const string& name) {
    auto ent = entity();
    // this lets you look up an entity by name
    $name_ents.insert_or_assign(name, ent);
    // this makes it easier to get the name during querying
    set<CellName>(ent, {name});
    return ent;
  }

  Entity UI::entity(const string& name) {
    dbc::check($name_ents.contains(name),
        fmt::format("GUECS entity {} does not exist. Mispelled cell name?", name));
    return $name_ents.at(name);
  }

  Entity UI::entity(const string& name, int id) {
    return entity(fmt::format("{}{}", name, id));
  }

  void UI::init() {
    query<Background>([](auto, auto& bg) {
        bg.init();
    });

    query<lel::Cell, Rectangle>([](auto, auto& cell, auto& rect) {
      rect.init(cell);
    });

    query<lel::Cell, Effect>([](auto, auto& cell, auto& shader) {
      shader.init(cell);
    });

    query<Rectangle, Meter>([](auto, auto& bg, auto &meter) {
      bg.shape->setFillColor(meter.color);
    });

    query<lel::Cell, Meter>([](auto, auto &cell, auto& meter) {
      meter.init(cell);
    });

    query<lel::Cell, Textual>([this](auto, auto& cell, auto& text) {
      text.init(cell, $font);
    });

    query<lel::Cell, Label>([this](auto, auto& cell, auto& text) {
      text.init(cell, $font);
    });

    query<lel::Cell, Sprite>([&](auto, auto &cell, auto &sprite) {
      sprite.init(cell);
    });
  }

  void UI::debug_layout(sf::RenderWindow& window) {
    query<lel::Cell>([&](const auto, auto &cell) {
      sf::RectangleShape rect{{float(cell.w), float(cell.h)}};
      rect.setPosition({float(cell.x), float(cell.y)});
      rect.setFillColor(sf::Color::Transparent);
      rect.setOutlineColor(sf::Color::Red);
      rect.setOutlineThickness(2.0f);
      window.draw(rect);
    });
  }

  void UI::render(sf::RenderWindow& window) {
    if(auto bg = get_if<Background>(MAIN)) {
      window.draw(*(bg->shape));
    }

    query<Effect>([&](auto, auto& shader) {
      if(shader.$active) shader.step();
    });

    query<Rectangle>([&](auto ent, auto& rect) {
        render_helper(window, ent, true, rect.shape);
    });

    query<lel::Cell, Meter>([&](auto ent, auto& cell, auto &meter) {
      meter.render(cell);
      render_helper(window, ent, true, meter.bar.shape);
    });

    query<Sprite>([&](auto ent, auto& sprite) {
        render_helper(window, ent, false, sprite.sprite);
    });

    query<Label>([&](auto ent, auto& text) {
        render_helper(window, ent, false, text.text);
    });

    query<Textual>([&](auto ent, auto& text) {
        render_helper(window, ent, true, text.text);
    });
  }

  bool UI::mouse(float x, float y, bool hover) {
    int action_count = 0;

    query<lel::Cell>([&](auto ent, auto& cell) {
      if((x >= cell.x && x <= cell.x + cell.w) &&
          (y >= cell.y && y <= cell.y + cell.h))
      {
        do_if<Effect>(ent, [hover](auto& effect) {
          effect.$shader->setUniform("hover", hover);
          effect.run();
        });

        do_if<Sound>(ent, [hover](auto& sound) {
            // here set that it played then only play once
            sound.play(hover);
        });

        if(hover) return;

        click_on(ent);

        action_count++;
      } else {
        do_if<Effect>(ent, [hover](auto& effect) {
          effect.stop();
        });

        do_if<Sound>(ent, [hover](auto& sound) {
            // here set that it played then only play once
            sound.stop(hover);
        });
      }
    });

    return action_count > 0;
  }

  void UI::show_sprite(const string& region, const string& sprite_name) {
    auto ent = entity(region);

    if(auto sprite = get_if<Sprite>(ent)) {
      sprite->update(sprite_name);
    } else {
      set_init<Sprite>(ent, {sprite_name});
    }
  }

  void UI::show_text(const string& region, const wstring& content) {
    auto ent = entity(region);

    if(auto tc = get_if<Textual>(ent)) {
      tc->text->setString(content);
    } else {
      auto &cell = cell_for(ent);
      Textual to_set{content, THEME.TEXT_SIZE};
      to_set.init(cell, $font);
      set<Textual>(ent, to_set);
    }
  }

  void UI::click_on(const string& name, bool required) {
    auto ent = entity(name);

    if(required) {
      dbc::check(has<Clickable>(ent),
          fmt::format("click_on required '{}' to exist but it doesn't", name));
    }

    click_on(ent);
  }

  void UI::click_on(Entity ent) {
    if(auto clicked = get_if<Clickable>(ent)) {
      if(auto action_data = get_if<ActionData>(ent)) {
        clicked->action(ent, action_data->data);
      } else {
        clicked->action(ent, {});
      }
    }
  }

  void UI::show_label(const string& region, const wstring& content) {
    auto ent = entity(region);

    if(auto tc = get_if<Label>(ent)) {
      tc->text->setString(content);
    } else {
      auto &cell = cell_for(ent);
      Label to_set{content, THEME.LABEL_SIZE};
      to_set.init(cell, $font);
      to_set.text->setFillColor(THEME.TEXT_COLOR);
      set<Label>(ent, to_set);
    }
  }

  wstring to_wstring(const string& str) {
    std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>> $converter;
    return $converter.from_bytes(str);
  }
}
